==> genome-spy-0.7.0.tgz: package/src/app/genomeSpyApp.js <==
import lzString from "lz-string";

import GenomeSpy from "../genomeSpy";
import "../styles/genome-spy-app.scss";
import favIcon from "../img/genomespy-favicon.svg";
import { html, render } from "lit";

import { VISIT_STOP } from "../view/view";
import SampleView from "../view/sampleView/sampleView";
import BookmarkDatabase from "../sampleHandler/bookmarkDatabase";
import { asArray } from "../utils/arrayUtils";

import "../sampleHandler/provenanceToolbar-wc";
import "../sampleHandler/bookmarkButton-wc";
import "./toolbar-wc";
import { createRef, ref } from "lit/directives/ref";

/**
 * A simple wrapper for the GenomeSpy core.
 */
export default class GenomeSpyApp {
    /**
     *
     * @param {HTMLElement} appContainerElement
     * @param {import("../spec/root").RootSpec} config
     * @param {import("../options").EmbedOptions} options
     */
    constructor(appContainerElement, config, options = {}) {
        this.config = config;

        this.toolbarRef = createRef();

        this.appContainer = appContainerElement;
        if (this.isFullPage()) {
            this.appContainer.style.margin = "0";
            this.appContainer.style.padding = "0";
            this.appContainer.style.overflow = "hidden";

            setFavicon(favIcon);
        } else {
            this.appContainer.style.position = "relative";
        }

        // eslint-disable-next-line consistent-this
        const self = this;

        this.bookmarkDatabase =
            typeof config.specId == "string"
                ? new BookmarkDatabase(config.specId)
                : undefined;

        this._renderTemplate = () => {
            render(getAppBody(), self.appContainer);
        };

        function getAppBody() {
            return html`
                <div class="genome-spy-app">
                    <genome-spy-toolbar
                        ${ref(self.toolbarRef)}
                        .app=${self}
                    ></genome-spy-toolbar>
                    <div class="genome-spy-container"></div>
                </div>
            `;
        }

        /** @param {string} className */
        const elem = (className) =>
            /** @type {HTMLElement} */ (
                this.appContainer.getElementsByClassName(className)[0]
            );

        this._renderTemplate();

        elem("genome-spy-container").addEventListener("click", (event) => {
            elem("search-input").blur();
        });

        this.genomeSpy = new GenomeSpy(
            elem("genome-spy-container"),
            this.config,
            options
        );
    }

    toggleFullScreen() {
        if (!document.fullscreenElement) {
            this.appContainer.requestFullscreen();
        } else if (document.exitFullscreen) {
            document.exitFullscreen();
        }
    }

    isFullPage() {
        return this.appContainer == document.body;
    }

    async launch() {
        const result = await this.genomeSpy.launch();
        if (!result) {
            return;
        }

        this._replayProvenanceFromUrl();
        this.getSampleHandler()?.provenance.addListener(() => {
            this._updateUrl();
        });

        const toolbar = /** @type {import("./toolbar-wc").default} */ (
            this.toolbarRef.value
        );
        // Just trigger re-render. Need a way to broadcast this to all components.
        toolbar.appInitialized = true;

        const title = asArray(this.genomeSpy.spec.description ?? []);

        if (this.isFullPage() && title.length > 0) {
            document.title = "GenomeSpy - " + title;
        }
    }

    /**
     * Update provenance to url
     */
    _updateUrl() {
        const history = this.getSampleHandler().provenance.getActionHistory();

        // TODO: Test https://github.com/101arrowz/fflate as a replacement for lzString

        let hash = "";
        if (history.length) {
            hash =
                "#" +
                lzString.compressToEncodedURIComponent(JSON.stringify(history));
        }

        window.history.replaceState(
            undefined,
            document.title,
            window.location.pathname + window.location.search + hash
        );
    }

    _replayProvenanceFromUrl() {
        if (!this.getSampleHandler()) {
            return;
        }

        const hash = window.location.hash;
        if (hash && hash.length > 0) {
            const history = JSON.parse(
                lzString.decompressFromEncodedURIComponent(hash.substr(1))
            );

            if (Array.isArray(history)) {
                this.getSampleHandler().dispatchBatch(history);
            }
        }
    }

    /**
     *
     * @param {import("../spec/root").RootSpec} config
     */
    async updateConfig(config) {
        // TODO: provenance etc must be re-registered etc
        throw new Error("Broken");

        /*
        this.config = config;
        // TODO: Preserve viewport
        this.genomeSpy.destroy();
        await this.launch();
		*/
    }

    getSampleView() {
        if (!this.genomeSpy?.viewRoot) {
            return;
        }

        /** @type {import("../view/sampleView/sampleView").default} */
        let sampleView;

        this.genomeSpy.viewRoot.visit((view) => {
            if (view instanceof SampleView) {
                sampleView = view;
                return VISIT_STOP;
            }
        });

        return sampleView;
    }

    getSampleHandler() {
        return this.getSampleView()?.sampleHandler;
    }
}

/**
 * https://spemer.com/articles/set-favicon-with-javascript.html
 *
 * @param {string} favImg
 */
function setFavicon(favImg) {
    let headTitle = document.querySelector("head");
    let setFavicon = document.createElement("link");
    setFavicon.setAttribute("rel", "shortcut icon");
    setFavicon.setAttribute("href", favImg);
    headTitle.appendChild(setFavicon);
}

==> genome-spy-0.9.0.tgz: package/src/app/genomeSpyApp.js <==
import {
    compressToEncodedURIComponent,
    decompressFromEncodedURIComponent,
} from "lz-string";

import GenomeSpy from "../genomeSpy";
import "../styles/genome-spy-app.scss";
import favIcon from "../img/genomespy-favicon.svg";
import { html, render } from "lit";

import { VISIT_STOP } from "../view/view";
import SampleView from "../view/sampleView/sampleView";
import BookmarkDatabase from "../sampleHandler/bookmarkDatabase";
import { asArray } from "../utils/arrayUtils";

import "../sampleHandler/provenanceToolbar-wc";
import "../sampleHandler/bookmarkButton-wc";
import "./toolbar-wc";
import { createRef, ref } from "lit/directives/ref.js";
import { debounce } from "../utils/debounce";

/**
 * A simple wrapper for the GenomeSpy core.
 */
export default class GenomeSpyApp {
    /**
     *
     * @param {HTMLElement} appContainerElement
     * @param {import("../spec/root").RootSpec} config
     * @param {import("../options").EmbedOptions} options
     */
    constructor(appContainerElement, config, options = {}) {
        this.config = config;

        this.toolbarRef = createRef();

        this.appContainer = appContainerElement;
        if (this.isFullPage()) {
            this.appContainer.style.margin = "0";
            this.appContainer.style.padding = "0";
            this.appContainer.style.overflow = "hidden";

            setFavicon(favIcon);
        } else {
            this.appContainer.style.position = "relative";
        }

        // eslint-disable-next-line consistent-this
        const self = this;

        this.bookmarkDatabase =
            typeof config.specId == "string"
                ? new BookmarkDatabase(config.specId)
                : undefined;

        this._renderTemplate = () => {
            render(getAppBody(), self.appContainer);
        };

        function getAppBody() {
            return html`
                <div class="genome-spy-app">
                    <genome-spy-toolbar
                        ${ref(self.toolbarRef)}
                        .app=${self}
                    ></genome-spy-toolbar>
                    <div class="genome-spy-container"></div>
                </div>
            `;
        }

        /** @param {string} className */
        const elem = (className) =>
            /** @type {HTMLElement} */ (
                this.appContainer.getElementsByClassName(className)[0]
            );

        this._renderTemplate();

        elem("genome-spy-container").addEventListener("click", (event) => {
            elem("search-input").blur();
        });

        this.genomeSpy = new GenomeSpy(
            elem("genome-spy-container"),
            this.config,
            options
        );
    }

    toggleFullScreen() {
        if (!document.fullscreenElement) {
            this.appContainer.requestFullscreen();
        } else if (document.exitFullscreen) {
            document.exitFullscreen();
        }
    }

    isFullPage() {
        return this.appContainer == document.body;
    }

    async launch() {
        const result = await this.genomeSpy.launch();
        if (!result) {
            return;
        }

        await this._restoreStateFromUrl();
        this.getSampleHandler()?.provenance.addListener(() => {
            this._updateStateToUrl();
        });

        const debouncedUpdateUrl = debounce(
            () => this._updateStateToUrl(),
            500,
            false
        );
        for (const [, res] of this.genomeSpy.getNamedScaleResolutions()) {
            if (res.isZoomable()) {
                res.addEventListener("domain", debouncedUpdateUrl);
            }
        }

        const toolbar = /** @type {import("./toolbar-wc").default} */ (
            this.toolbarRef.value
        );
        // Just trigger re-render. Need a way to broadcast this to all components.
        toolbar.appInitialized = true;

        const title = asArray(this.genomeSpy.spec.description ?? []);

        if (this.isFullPage() && title.length > 0) {
            document.title = "GenomeSpy - " + title;
        }
    }

    /**
     * Update provenance to url
     */
    _updateStateToUrl() {
        /** @type {import("./genomeSpyAppTypes").UrlHash} */
        const hashData = {
            actions: [],
            scaleDomains: {},
        };

        const history = this.getSampleHandler()?.provenance.getActionHistory();
        if (history?.length) {
            hashData.actions = history;
        }

        // This is copypaste from bookmarks. TODO: consolidate
        for (const [name, scaleResolution] of this.genomeSpy
            .getNamedScaleResolutions()
            .entries()) {
            if (!scaleResolution.isZoomed()) {
                hashData.scaleDomains[name] =
                    scaleResolution.getComplexDomain();
            }
        }

        // TODO: Test https://github.com/101arrowz/fflate as a replacement for lzString

        let hash =
            hashData.actions.length || Object.keys(hashData.scaleDomains).length
                ? "#" + compressToEncodedURIComponent(JSON.stringify(hashData))
                : "";

        window.history.replaceState(
            undefined,
            document.title,
            window.location.pathname + window.location.search + hash
        );
    }

    async _restoreStateFromUrl() {
        const hash = window.location.hash;
        if (hash && hash.length > 0) {
            try {
                /** @type {import("./genomeSpyAppTypes").UrlHash} */
                const hashData = JSON.parse(
                    decompressFromEncodedURIComponent(hash.substr(1))
                );

                const sampleHandler = this.getSampleHandler();
                if (sampleHandler && hashData.actions) {
                    // This is copypaste from bookmarks. TODO: consolidate

                    sampleHandler.provenance.activateState(0);
                    sampleHandler.dispatchBatch(hashData.actions);
                }

                /** @type {Promise<void>[]} */
                const promises = [];
                for (const [name, scaleDomain] of Object.entries(
                    hashData.scaleDomains ?? {}
                )) {
                    const scaleResolution = this.genomeSpy
                        .getNamedScaleResolutions()
                        .get(name);
                    if (scaleResolution) {
                        // @ts-ignore
                        promises.push(scaleResolution.zoomTo(scaleDomain));
                    } else {
                        console.warn(
                            `Cannot restore scale domain. Unknown name: ${name}`
                        );
                    }
                }
                await Promise.all(promises);
            } catch (e) {
                console.error(e);
                alert(`Cannot restore state from URL:\n${e}`);
                this.getSampleHandler()?.provenance.activateState(0);
            }
        }
    }

    /**
     *
     * @param {import("../spec/root").RootSpec} config
     */
    async updateConfig(config) {
        // TODO: provenance etc must be re-registered etc
        throw new Error("Broken");

        /*
        this.config = config;
        // TODO: Preserve viewport
        this.genomeSpy.destroy();
        await this.launch();
		*/
    }

    getSampleView() {
        if (!this.genomeSpy?.viewRoot) {
            return;
        }

        /** @type {import("../view/sampleView/sampleView").default} */
        let sampleView;

        this.genomeSpy.viewRoot.visit((view) => {
            if (view instanceof SampleView) {
                sampleView = view;
                return VISIT_STOP;
            }
        });

        return sampleView;
    }

    getSampleHandler() {
        return this.getSampleView()?.sampleHandler;
    }
}

/**
 * https://spemer.com/articles/set-favicon-with-javascript.html
 *
 * @param {string} favImg
 */
function setFavicon(favImg) {
    let headTitle = document.querySelector("head");
    let setFavicon = document.createElement("link");
    setFavicon.setAttribute("rel", "shortcut icon");
    setFavicon.setAttribute("href", favImg);
    headTitle.appendChild(setFavicon);
}
